/* 
   Changes 3/13/2000:
      + We now print the min/max/avg min_loc/max_loc for the communication
        partners across all ranks, instead of printing the list of all 
	ranks which participated in a point-to-point communication.
        In a 5-pt stencil, for instance, the avg. should be close to 5
	with min/max values differing due to boundary effects.
	The heading for this section (in MPI_PROFILE_FILE) has changed
	to "Communication Partners".

   Changes 3/3/2000:
      + On Linux systems, the /proc/(pid)/stat file is read in order
        to get memory usage info. Lines are notated with a 
	#ifdef LINUX directive.
      + We now check for the T5_EVENT1 environment variable, since
        if this is set, it generates an error (conflicting hwc error)
	on SGI machines when using the hardware counters. If set, an error
	message is printed to STDERR and executions halts.
	If not set, we check for the presence of the MPI_PROFILE_EVENT1
	environment variable. If set to 21, floating point ops are polled 
	from the hardware counters, and if set to 26,  L2 cache misses 
	are polled. Other values are allowed since we don't check this
	value. If MPI_PROFILE_EVENT1 is not set, L2 cache misses are
	polled by default.
      + The destination/sources of each point-to-point communication
        call is tracked and printed out, under the heading
	of "Processor Communication Data". 
	We just keep track of the ranks, independent of the communicators.  
	We *should* use MPI_Group_translate_ranks to make all ranks 
	relative to MPI_COMM_WORLD; in another realease...

      o Still to do: The Fortran interfaces for the MI_Waitxxx routines
	are incorrect and need to be fixed.

   Changes 2/24/2000:
      + Totals for point-to-point sends:
          total calls, size of messages, time. This includes
	  the usual min/max/avg min-loc/max-loc info as well.
	  Also, the communication rate (size/time) is calculated.
      + Total for collective sends: (same as above)
      + Total "barrier" time is shown for collective routines
        (this is the time spent waiting for nodes to initiate 
        collective sends/receives)

   Changes 2/22/2000:
      + added profiling for MPI_Waitxxx calls (C only)
      + stats printed for Waits & Barrier calls.
      + added more MPI routines (Sendrecv_replace, Pack/Unpack)
*/

#include "profiler.h"
#include "mpi.h"
#include <fcntl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/resource.h>
#include <time.h>
#include <unistd.h>

/* Global defs & variables */

#define NBINS 29
#define ONE_MB 1048576
#define ONE_KB 1024
#define MAX_MSG 1073741824
#define MAX_PROCBY32 32    /* This allows for 2^32 ~ 4.2 billion processors */
#define MPI_PROF_FILE_NAME "profile.txt"

#define MPI_PROF_FLOPS 21
#define MPI_PROF_L2 26

#ifdef HAVE_LIBPERFEX
/* R10K hardware performance counters, these can be overridden using
   the environment variables T5_EVENT0 and T5_EVENT1
   Note: If T5_EVENT1 is set, a 'conflicting events' message is generated,
         regardless if it matches the defs below.
 */
static int hwpc_e0 = 0;              /* cycles */
static int  hwpc_e1 = MPI_PROF_L2;   /* L2 cache misses (default) */
static long long hwpc_c0, hwpc_c1;
static int hwpc_ok = 1;
#endif

#ifdef HAVE_LIBPCL
#include <pcl.h>

/* PCL Variables */
#define PCL_NEVENTS 2
int counter_list[PCL_NEVENTS];     /* A list of events to track */
int ncounter, PCL_NO_EVENTS = 0;
unsigned int mode;
PCL_CNT_TYPE i_result_list[PCL_NEVENTS];
PCL_FP_CNT_TYPE fp_result_list[PCL_NEVENTS];
#endif

/* timing */
static clock_t MPI_Start_time, MPI_End_time;

static double startwtime, endwtime; /* These get overwritten as needed */
static double sync_time;            /* also used as needed */

/* Program name */
static char *cmdname = 0;

/* Message size bins */
static int bin_id;
enum bin_sizes {sz_0, sz_1to4, sz_5to8, sz_9to16, sz_17to32, sz_33to64, 
		sz_65to128, sz_129to256, sz_257to512, sz_513to1024, 
		sz_1to4K, sz_5to8K, sz_9to16K, sz_17to32K, sz_33to64K, 
		sz_65to128K, sz_129to256K, sz_257to512K, sz_513to1024K,
		sz_1to4M, sz_5to8M, sz_9to16M, sz_17to32M, sz_33to64M,
		sz_65to128M, sz_129to256M, sz_257to512M, sz_513to1024M, 
		sz_large};

static char *bin_names[NBINS] =  {"0 ", "1 to 4 ", "5 to 8 ", "9 to 16 ", 
			   "17 to 32 ", "33 to 64 ", "65 to 128 ", 
			   "129 to 256 ", "257 to 512 ", "513 to 1024 ", 
			   "1 to 4 K", "5 to 8 K", "9 to 16 K",
			   "17 to 32 K", "33 to 64 K", "65 to 128 K", 
			   "129 to 256 K", "257 to 512 K", "513 to 1024 K",
			   "1 to 4 M", "5 to 8 M", "9 to 16 M", "17 to 32 M", 
			   "33 to 64 M", "65 to 128 M", "129 to 256 M", 
			   "257 to 512 M", "513 to 1024 M", "> 1024 M"};

#if defined(NCSAPERF)
static char *max_bin_names[NBINS] =  
      {"0B ", "0-4B ", "4-8B ", "8-16B ", "16-32B ",
       "32-64B ", "64-128B ", "128-256B ", "256-512B ", 
       "512-1023B ", "1024B-4KB ", "4-8KB ", "8-16KB ",
       "16-32KB ", "32-64KB ", "64-128KB ",
       "128-256KB ", "256-512KB ", "512-1024KB ",
       "1024KB-4MB ", "4-8MB ", "8-16MB ",
       "16-32MB ", "32-64MB ", "64-128MB",
       "128-256MB ", "256-512MB ", "512-1024MB ", 
       ">1024MB "};
#else
static char *max_bin_names[NBINS] =  
      {"0 ", "4 ", "8 ", "16 ", "32 ", "64 ", "128 ", 
       "256 ", "512 ", "1024 ", "4 K", "8 K", 
       "16 K","32 K", "64 K", "128 K", "256 K", 
       "512 K", "1024 K", "4 M", "8 M", "16 M", 
       "32 M", "64 M", "128 M", "256 M", "512 M", 
       "1024 M", "> 1024 M"};
#endif

#define COUNTER_DECL(name) \
    static int name##_calls[NBINS]; \
    static long name##_size[NBINS]; \
    static double name##_time[NBINS]

#define COUNTER_DECL_COLL(name) \
    COUNTER_DECL(name); \
    static double name##_timesync[NBINS]

COUNTER_DECL_COLL(allgather);
COUNTER_DECL_COLL(allgatherv);
COUNTER_DECL_COLL(allreduce);
COUNTER_DECL_COLL(alltoall);
COUNTER_DECL_COLL(alltoallv);
/* COUNTER_DECL(barrier) barrier is special */

static int    barrier_calls = 0;
static int    wait_calls = 0;
static int    waitall_calls = 0;
static int    waitany_calls = 0;
static int    waitsome_calls = 0;

static double barrier_time = 0;
static double wait_time = 0;
static double waitall_time = 0;
static double waitany_time = 0;
static double waitsome_time = 0;

#ifdef HAVE_PCL
PCL_CNT_TYPE i_result_list[PCL_NEVENTS];
PCL_FP_CNT_TYPE fp_result_list[PCL_NEVENTS];
#endif

COUNTER_DECL(bcast);
COUNTER_DECL(gather);
COUNTER_DECL(gatherv);
COUNTER_DECL(reduce_scatter);
COUNTER_DECL(reduce);
COUNTER_DECL(scan);
COUNTER_DECL(scatter);
COUNTER_DECL(scatterv);

COUNTER_DECL(recv);
COUNTER_DECL(send);
COUNTER_DECL(sendrecv);
COUNTER_DECL(sendrecv_replace);
COUNTER_DECL(bsend);
COUNTER_DECL(ibsend);
COUNTER_DECL(isend);
COUNTER_DECL(issend);
COUNTER_DECL(irsend);
COUNTER_DECL(rsend);
COUNTER_DECL(ssend);
COUNTER_DECL(irecv);

COUNTER_DECL(pack);
COUNTER_DECL(unpack);

static int    total_isend_calls = 0;
static long   total_isend_size = 0;
static double total_isend_time = 0;
static int    total_bsend_calls = 0;
static long   total_bsend_size = 0;
static double total_bsend_time = 0;
static int    total_col_send_calls = 0;
static long   total_col_send_size = 0;
static double total_col_send_time = 0;

/* Bit vectors to keep track of which processes this one has been in
   communication with. */
/* On Linux (gcc) ints are 32-bit */
#define COMM_DESTS(name) \
    static int32 name##_target[MAX_PROCBY32]

COMM_DESTS(recv);
COMM_DESTS(send);
COMM_DESTS(sendrecv);
COMM_DESTS(sendrecv_replace);
COMM_DESTS(bsend);
COMM_DESTS(ibsend);
COMM_DESTS(isend);
COMM_DESTS(issend);
COMM_DESTS(irsend);
COMM_DESTS(rsend);
COMM_DESTS(ssend);
COMM_DESTS(irecv);

#define UPDATE_COMM_DESTS(name,source) \
    name##_target[source >> 5] |= (1 << (source & 0x01F))

#define Bit_Is_Set(target,SelectBit) \
( (target[i] & SelectBit) > 0 )

/* forward declarations */
int32 UArraySum(int32 *, int);
static int Get_Bin_ID( int );
static int Get_Msg_size( int, MPI_Datatype );
static void printout_barstats( int , double, const char *, FILE *, int, int);
static void printout_header( const char *, FILE *, int );
static void printout_msgstats_coll( int [], long [], double [], double[], 
			       const char *, FILE *, int, int );
static void printout_total_msgstats( FILE *, int, int);
static void printout_comm_dests( int32 *, const char *, FILE * );

static void getrange_int( const int[], int, 
			  int [], int [], int [], int [], int [] );
static void getrange_long( const long[], int, long [], long [], long [], 
			   int [], int [] ); 
static void getrange_double( const double[], int, 
			     double [], double [], double [], int [], int [] );

static void getrange_int_single(int, int *, int *,int *, int *, int *t);
static void getrange_double_single(double, double *, int *, double *, 
				   int *, double *);

#ifdef HAVE_PCL
static void printout_pclstats( int, int, double, FILE *, int, int);
#endif

#ifdef LINUX
int ReadProcStat( procstat_t *); /* Reads the /proc/(pid)/stat file */
#endif

/* Begin Profiling Routines */

/* This macro lets us change the way we remember the call */
#define UPDATE_COUNTER(name) \
    name##_time[bin_id] += endwtime - startwtime; \
    name##_calls[bin_id] ++; \
    name##_size[bin_id] += buf_size

#define UPDATE_COUNTER_COLL(name) \
    UPDATE_COUNTER(name);\
    name##_timesync[bin_id] += sync_time

/*
    MPI_Allgather - prototyping replacement for MPI_Allgather
    Trace the beginning and ending of MPI_Allgather.
*/
int MPI_Allgather( void *sendbuf, int sendcount, MPI_Datatype sendtype, 
		   void *recvbuf, int recvcount, MPI_Datatype recvtype, 
		   MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( sendcount, sendtype );
  bin_id = Get_Bin_ID( buf_size );

  startwtime = MPI_Wtime();
  PMPI_Barrier( comm );
  sync_time = MPI_Wtime() - startwtime;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Allgather( sendbuf, sendcount, sendtype, recvbuf, 
			      recvcount, recvtype, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER_COLL(allgather);

  return returnVal;
}

/*
    MPI_Allgatherv - prototyping replacement for MPI_Allgatherv
    Trace the beginning and ending of MPI_Allgatherv.
*/
int MPI_Allgatherv( void *sendbuf, int sendcount, MPI_Datatype sendtype, 
		    void *recvbuf, int *recvcounts, int *displs, 
		    MPI_Datatype recvtype, MPI_Comm comm ) 
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( sendcount, sendtype );
  bin_id = Get_Bin_ID( buf_size);

  startwtime = MPI_Wtime();
  PMPI_Barrier( comm );
  sync_time = MPI_Wtime() - startwtime;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Allgatherv( sendbuf, sendcount, sendtype, recvbuf, 
			       recvcounts, displs, recvtype, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER_COLL(allgatherv);
  
  return returnVal;
}

/*
    MPI_Allreduce - prototyping replacement for MPI_Allreduce
    Trace the beginning and ending of MPI_Allreduce.
*/
int MPI_Allreduce( void *sendbuf, void *recvbuf, int count, 
		   MPI_Datatype datatype, MPI_Op op, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID( buf_size );

  startwtime = MPI_Wtime();
  PMPI_Barrier( comm );
  sync_time = MPI_Wtime() - startwtime;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Allreduce( sendbuf, recvbuf, count, datatype, op, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER_COLL(allreduce);

  return returnVal;
}

/*
    MPI_Alltoall - prototyping replacement for MPI_Alltoall
    Trace the beginning and ending of MPI_Alltoall.
*/
int MPI_Alltoall( void *sendbuf, int sendcount, MPI_Datatype sendtype, 
		  void *recvbuf, int recvcnt, MPI_Datatype recvtype, 
		  MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( sendcount, sendtype );
  bin_id = Get_Bin_ID( buf_size );

  startwtime = MPI_Wtime();
  PMPI_Barrier( comm );
  sync_time = MPI_Wtime() - startwtime;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Alltoall( sendbuf, sendcount, sendtype, recvbuf,
			     recvcnt, recvtype, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER_COLL(alltoall);
  
  return returnVal;
}

/*
    MPI_Alltoallv - prototyping replacement for MPI_Alltoallv
    Trace the beginning and ending of MPI_Alltoallv.
*/
int MPI_Alltoallv( void *sendbuf, int *sendcnts, int *sdispls, 
		   MPI_Datatype sendtype, void *recvbuf, int *recvcnts, 
		   int *rdispls, MPI_Datatype recvtype, MPI_Comm comm )
{
  int returnVal;
  int buf_size, i, nprocs;

  buf_size = *sendcnts * sizeof(sendtype);
  buf_size = 0;
  PMPI_Comm_size( comm, &nprocs );
  for (i=0; i<nprocs; i++) {
      buf_size += Get_Msg_size( sendcnts[i], sendtype );
  }
  bin_id = Get_Bin_ID(buf_size);  /* Determine message bin ID */

  startwtime = MPI_Wtime();
  PMPI_Barrier( comm );
  sync_time = MPI_Wtime() - startwtime;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Alltoallv( sendbuf, sendcnts, sdispls, sendtype, recvbuf,
			      recvcnts, rdispls, recvtype, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER_COLL(alltoallv);
  
  return returnVal;
}

/*
    MPI_Barrier - prototyping replacement for MPI_Barrier
    Trace the beginning and ending of MPI_Barrier.
*/
int MPI_Barrier( MPI_Comm comm )
{
  int returnVal;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Barrier( comm );
  endwtime = MPI_Wtime();
  barrier_time  += endwtime - startwtime;
  barrier_calls += 1;

  return returnVal;
}

/*
    MPI_Bcast - prototyping replacement for MPI_Bcast
    Trace the beginning and ending of MPI_Bcast.
*/
int MPI_Bcast( void *buffer, int count, MPI_Datatype datatype, 
	       int root, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );/* Size is in bytes */
  bin_id = Get_Bin_ID(buf_size);  /* Determine message bin ID */

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Bcast( buffer, count, datatype, root, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER(bcast);

  return returnVal;
}

/*
    MPI_Gather - prototyping replacement for MPI_Gather
    Trace the beginning and ending of MPI_Gather.
*/
int MPI_Gather( void *sendbuf, int sendcnt, MPI_Datatype sendtype, 
		void *recvbuf, int recvcount, MPI_Datatype recvtype, 
		int root, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( sendcnt, sendtype );
  bin_id = Get_Bin_ID(buf_size);
  
  startwtime = MPI_Wtime();
  returnVal = PMPI_Gather( sendbuf, sendcnt, sendtype, recvbuf, 
			   recvcount, recvtype, root, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER(gather);

  return returnVal;
}

/*
    MPI_Gatherv - prototyping replacement for MPI_Gatherv
    Trace the beginning and ending of MPI_Gatherv.
*/
int MPI_Gatherv( void *sendbuf, int sendcnt, MPI_Datatype sendtype, 
		 void *recvbuf, int *recvcnts, int *displs, 
		 MPI_Datatype recvtype, int root, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( sendcnt, sendtype );
  bin_id = Get_Bin_ID(buf_size);
  
  startwtime = MPI_Wtime();
  returnVal = PMPI_Gatherv( sendbuf, sendcnt, sendtype, recvbuf, recvcnts,
			    displs, recvtype, root, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER(gatherv);
  
  return returnVal;
}

/*
    MPI_Pack - prototyping replacement for MPI_Pack
    Log the beginning and ending of the time spent in MPI_Pack calls.
*/
int MPI_Pack( void *inbuf, int incount, MPI_Datatype datatype, void *outbuf, 
	      int outcount, int *position, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( incount, datatype );
  bin_id = Get_Bin_ID( buf_size );
  
  startwtime = MPI_Wtime();
  returnVal  = PMPI_Pack(inbuf, incount, datatype, outbuf, outcount, 
			position, comm );
  endwtime   = MPI_Wtime();
  UPDATE_COUNTER(pack);

  return returnVal;
}

/*
    MPI_Unpack - prototyping replacement for MPI_Unpack
    Log the beginning and ending of the time spent in MPI_Unpack calls.
*/
int MPI_Unpack( void *inbuf, int insize, int *position, void *outbuf, 
		int outcount, MPI_Datatype datatype, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( outcount, datatype );
  bin_id = Get_Bin_ID( buf_size );
  
  startwtime = MPI_Wtime();
  returnVal = PMPI_Unpack( inbuf, insize, position, outbuf, outcount, 
			   datatype, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER(unpack);

  return returnVal;
}

/*
    MPI_Reduce_scatter - prototyping replacement for MPI_Reduce_scatter
    Trace the beginning and ending of MPI_Reduce_scatter.
*/
int MPI_Reduce_scatter( void *sendbuf, void *recvbuf, int *recvcnts, 
			MPI_Datatype datatype, MPI_Op op, MPI_Comm comm )
{
  int returnVal;
  int ierr, buf_size;
  
  buf_size = Get_Msg_size( recvcnts[0], datatype ); /* FIXME */
  bin_id = Get_Bin_ID(buf_size);

  startwtime	= MPI_Wtime();	  
  ierr		= PMPI_Barrier( comm );
  endwtime	= MPI_Wtime();	  
  sync_time     = startwtime - endwtime;

  startwtime = endwtime;
  returnVal = PMPI_Reduce_scatter( sendbuf, recvbuf, recvcnts, datatype,
				   op, comm );
  endwtime = MPI_Wtime();	
  UPDATE_COUNTER(reduce_scatter);  
  
  return returnVal;
}

/*
    MPI_Reduce - prototyping replacement for MPI_Reduce
    Trace the beginning and ending of MPI_Reduce.
*/
int MPI_Reduce( void *sendbuf, void *recvbuf, int count, 
		  MPI_Datatype datatype, MPI_Op op, int root, 
		  MPI_Comm comm )
{
  int returnVal;
  int buf_size;
  
  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = endwtime;
  returnVal = PMPI_Reduce( sendbuf, recvbuf, count, datatype, op, root, comm );
  endwtime = MPI_Wtime();	
  UPDATE_COUNTER(reduce);  

  return returnVal;
}

/*
    MPI_Scan - prototyping replacement for MPI_Scan
    Trace the beginning and ending of MPI_Scan.
*/
int MPI_Scan( void *sendbuf, void *recvbuf, int count, 
		MPI_Datatype datatype, MPI_Op op, MPI_Comm comm )
{
  int returnVal;
  int buf_size;
  
  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = endwtime;
  returnVal = PMPI_Scan( sendbuf, recvbuf, count, datatype, op, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER(scan);	  

  return returnVal;
}

/*
    MPI_Scatter - prototyping replacement for MPI_Scatter
    Trace the beginning and ending of MPI_Scatter.
*/
int MPI_Scatter( void *sendbuf, int sendcnt, MPI_Datatype sendtype, 
		 void *recvbuf, int recvcnt, MPI_Datatype recvtype, 
		 int root, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( sendcnt, sendtype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = endwtime;
  returnVal = PMPI_Scatter( sendbuf, sendcnt, sendtype, recvbuf, recvcnt, 
			    recvtype, root, comm );
  endwtime = MPI_Wtime();	  
  UPDATE_COUNTER(scatter);

  return returnVal;
}

/*
    MPI_Scatterv - prototyping replacement for MPI_Scatterv
    Trace the beginning and ending of MPI_Scatterv.
*/
int MPI_Scatterv( void *sendbuf, int *sendcnts, int *displs, 
		  MPI_Datatype sendtype, void *recvbuf, int recvcnt, 
		  MPI_Datatype recvtype, int root, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( sendcnts[0], sendtype ); /* FIXME */
  bin_id = Get_Bin_ID(buf_size);

  startwtime = endwtime;
  returnVal = PMPI_Scatterv( sendbuf, sendcnts, displs, sendtype, recvbuf, 
			     recvcnt, recvtype, root, comm );
  endwtime = MPI_Wtime();	
  UPDATE_COUNTER(scatterv);  

  return returnVal;
}

/*
    MPI_Init - prototyping replacement for MPI_Init
    Trace the beginning and ending of MPI_Init.
*/
int MPI_Init( int *argc, char ***argv )
{
  int returnVal;
  int ierr, llrank;
  int MPI_PROF_EVENT1, event1;

  MPI_Start_time = clock();
  
  returnVal = PMPI_Init( argc, argv );

  if (argv) cmdname = (*argv)[0];

  PMPI_Comm_rank( MPI_COMM_WORLD, &llrank );

  /* If a user has the env variable T5_EVENT1 set, this code will generate
     a conflicting events error when trying to set the hardware counters
     We use a flag to see if the variable has been set. */
#ifdef HAVE_LIBPERFEX
  if (getenv("T5_EVENT1") != NULL) {
    fprintf(stderr, "T5_EVENT1 detected, aborting run.\n");
    returnVal = PMPI_Finalize();
    return returnVal;
  }
  else if (getenv("MPI_PROFILE_EVENT1") != NULL)
    hwpc_e1 = atoi( getenv("MPI_PROFILE_EVENT1") );

  if ( start_counters( hwpc_e0, hwpc_e1 ) < 0 ) 
    hwpc_ok = 0;
#endif

#ifdef HAVE_PCL
  /* Define PCL measurements */
  ncounter = PCL_NEVENTS;
  counter_list[0] = PCL_CYCLES;
  counter_list[1] = PCL_FP_INSTR;
  /* Define PCL count mode */
  mode = PCL_MODE_USER;

  /* Check if PCL works on this machine */
  ierr = PCLquery(counter_list, ncounter, mode);
  if ( ierr != PCL_SUCCESS) {
    fprintf(stderr,"[%d] Failed with error %d\n", llrank, ierr);
    if (ierr == PCL_NOT_SUPPORTED)
      fprintf(stderr, "[%d] Requested PCL event(s) not available.\n", llrank);
    else if (ierr == PCL_TOO_MANY_EVENTS)
      fprintf(stderr, "[%d] Too many requested events.\n", llrank);
    else if (ierr == PCL_ILL_EVENT)
      fprintf(stderr, "[%d] Illegal event identifier.\n", llrank);
    else if (ierr == PCL_MODE_NOT_SUPPORTED)
      fprintf(stderr, "[%d] Unsupported PCL MODE.\n", llrank);
    else if (ierr == PCL_FAILURE)
      fprintf(stderr, "[%d] Unknown PCL Failure.\n", llrank);
    return ierr;
  }

  /* Start performance counting */
  ierr = PCLstart(counter_list, ncounter, mode);
  if (ierr != PCL_SUCCESS) {
    fprintf( stderr, "[%d] PCLstart failed!\n", llrank);
  }
#endif
  
  return returnVal;
}

/*
    MPI_Recv - prototyping replacement for MPI_Recv
    Trace the beginning and ending of MPI_Recv.
*/
int MPI_Recv( void *buf, int count, MPI_Datatype datatype, int source, 
	       int tag, MPI_Comm comm, MPI_Status *status )
{
  int returnVal;
  int buf_size;
  
  startwtime = MPI_Wtime();
  returnVal = PMPI_Recv( buf, count, datatype, source, tag, comm, status );
  endwtime = MPI_Wtime();

  MPI_Get_count( status, datatype, &count );
  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);  /* Determine message bin ID */
  
  UPDATE_COUNTER(recv);
  UPDATE_COMM_DESTS(recv, source);

  return returnVal;
}

/*
    MPI_Send - prototyping replacement for MPI_Send
    Trace the beginning and ending of MPI_Send.
*/
int MPI_Send( void *buf, int count, MPI_Datatype datatype, int dest, 
	       int tag, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal  = PMPI_Send( buf, count, datatype, dest, tag, comm );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(send);
  UPDATE_COMM_DESTS(send, dest);

  return returnVal;
}

/*
    MPI_Sendrecv - prototyping replacement for MPI_Sendrecv
    Trace the beginning and ending of MPI_Sendrecv.
*/
int MPI_Sendrecv( void * sendbuf, int sendcount, MPI_Datatype sendtype, 
		   int dest, int sendtag, void * recvbuf, int recvcount, 
		   MPI_Datatype recvtype, int source, int recvtag, 
		   MPI_Comm comm, MPI_Status * status)
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( sendcount, sendtype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Sendrecv( sendbuf, sendcount, sendtype, dest, sendtag, 
			     recvbuf, recvcount, recvtype, source, recvtag,
			     comm, status );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(sendrecv);
  UPDATE_COMM_DESTS(sendrecv, dest);

  return returnVal;
}

/*
    MPI_Sendrecv_replace - prototyping replacement for MPI_Sendrecv_replace
    Log the beginning and ending of the time spent in MPI_Sendrecv_replace calls.
*/
int MPI_Sendrecv_replace( void *buf, int count, MPI_Datatype datatype, 
			  int dest, int sendtag, int source, int recvtag, 
			  MPI_Comm comm, MPI_Status *status )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal  = PMPI_Sendrecv_replace( buf, count, datatype, dest, 
			 sendtag, source, recvtag, comm, status );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(sendrecv_replace);
  UPDATE_COMM_DESTS(sendrecv_replace, dest);
  
  return returnVal;
}

/*
    MPI_Bsend - prototyping replacement for MPI_Bsend
    Trace the beginning and ending of MPI_Bsend.
*/
int MPI_Bsend( void *buf, int count, MPI_Datatype datatype, int dest, 
	       int tag, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Bsend( buf, count, datatype, dest, tag, comm );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(bsend);
  UPDATE_COMM_DESTS(bsend, dest);

  return returnVal;
}

/*
    MPI_Ibsend - prototyping replacement for MPI_Ibsend
    Trace the beginning and ending of MPI_Ibsend.
*/
int MPI_Ibsend( void *buf, int count, MPI_Datatype datatype, int dest, 
		int tag, MPI_Comm comm, MPI_Request *request )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);
  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Ibsend( buf, count, datatype, dest, tag, comm, request );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(ibsend);
  UPDATE_COMM_DESTS(ibsend, dest);

  return returnVal;
}

/*
    MPI_Irecv - prototyping replacement for MPI_Irecv
    Trace the beginning and ending of MPI_Irecv.
*/
int MPI_Irecv( void * buf, int count, MPI_Datatype datatype, int source,
		int tag, MPI_Comm comm, MPI_Request * request)
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);
  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Irecv( buf, count, datatype, source, tag, comm, request );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(irecv);
  UPDATE_COMM_DESTS(irecv, source);

  return returnVal;
}

/*
    MPI_Irsend - prototyping replacement for MPI_Irsend
    Trace the beginning and ending of MPI_Irsend.
*/
int MPI_Irsend( void *buf, int count, MPI_Datatype datatype, int dest, 
		int tag, MPI_Comm comm, MPI_Request *request )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Irsend( buf, count, datatype, dest, tag, comm, request );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(irsend);
  UPDATE_COMM_DESTS(irsend, dest);

  return returnVal;
}

/*
    MPI_Rsend - prototyping replacement for MPI_Rsend
    Trace the beginning and ending of MPI_Rsend.
*/
int MPI_Rsend( void * buf, int count, MPI_Datatype datatype, int dest,
	       int tag, MPI_Comm comm)
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Rsend( buf, count, datatype, dest, tag, comm );
  endwtime   = MPI_Wtime();	
  UPDATE_COUNTER(rsend);  
  UPDATE_COMM_DESTS(rsend, dest);

  return returnVal;
}

/*
    MPI_Ssend - prototyping replacement for MPI_Ssend
    Trace the beginning and ending of MPI_Ssend.
*/
int MPI_Ssend( void * buf, int count, MPI_Datatype datatype, int dest,
		int tag, MPI_Comm comm)
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Ssend( buf, count, datatype, dest, tag, comm );
  endwtime   = MPI_Wtime();	
  UPDATE_COUNTER(ssend);  
  UPDATE_COMM_DESTS(ssend, dest);

  return returnVal;
}
/*
    MPI_Isend - prototyping replacement for MPI_Isend
    Trace the beginning and ending of MPI_Isend.
*/
int MPI_Isend( void *buf, int count, MPI_Datatype datatype, int dest, 
		int tag, MPI_Comm comm, MPI_Request *request )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Isend( buf, count, datatype, dest, tag, comm, request );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(isend);
  UPDATE_COMM_DESTS(isend, dest);

  return returnVal;
}

/*
    MPI_Issend - prototyping replacement for MPI_Issend
    Trace the beginning and ending of MPI_Issend.
*/
int MPI_Issend( void * buf, int count, MPI_Datatype datatype, 
		 int dest, int tag, MPI_Comm comm, MPI_Request * request)
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Issend( buf, count, datatype, dest, tag, comm, request );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(issend);
  UPDATE_COMM_DESTS(issend, dest);

  return returnVal;
}

/*
    MPI_Wait - prototyping replacement for MPI_Wait
    Log the beginning and ending of the time spent in MPI_Wait calls.
*/
int MPI_Wait( MPI_Request *request, MPI_Status *status )
{
  int returnVal;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Wait( request, status );
  endwtime = MPI_Wtime();
  wait_time  += endwtime - startwtime;
  wait_calls += 1;

  return returnVal;
}

/*
    MPI_Waitall - prototyping replacement for MPI_Waitall
    Log the beginning and ending of the time spent in MPI_Waitall calls.
*/
int MPI_Waitall( int count, MPI_Request *array_of_requests, 
		 MPI_Status *array_of_statuses )
{
  int returnVal;
  
  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Waitall( count, array_of_requests, array_of_statuses );
  endwtime = MPI_Wtime();
  waitall_time  += endwtime - startwtime;
  waitall_calls += 1;

  return returnVal;
}

/*
    MPI_Waitany - prototyping replacement for MPI_Waitany
    Log the beginning and ending of the time spent in MPI_Waitany calls.
*/
int MPI_Waitany( int count, MPI_Request *array_of_requests, 
		 int *index, MPI_Status *status )
{
  int returnVal;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Waitany( count, array_of_requests, index, status );
  endwtime = MPI_Wtime();
  waitany_time  += endwtime - startwtime;
  waitany_calls += 1;

  return returnVal;
}

/*
    MPI_Waitsome - prototyping replacement for MPI_Waitsome
    Log the beginning and ending of the time spent in MPI_Waitsome calls.
*/
int MPI_Waitsome( int incount, MPI_Request *array_of_requests, 
		  int *outcount, int *array_of_indices, 
		  MPI_Status *array_of_statuses )
{
  int returnVal;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Waitsome( incount, array_of_requests, outcount, 
			     array_of_indices, array_of_statuses );
  endwtime = MPI_Wtime();
  waitsome_time  += endwtime - startwtime;
  waitsome_calls += 1;

  return returnVal;
}

/* Convert a (count/datatype) into a size in bytes */
static int Get_Msg_size( int count, MPI_Datatype datatype )
{
    int dsize;
    
    PMPI_Type_size( datatype, &dsize );
    return count * dsize;
}

/* Determine message bin ID.
   To keep this fast, the bin id's are a power of two; we can compute the
   bin id with a quick tree code.  STILL TO DO!  FIXME
*/
static int Get_Bin_ID(int buf_size)
{
  int bin_id = 0; /* By default, assume a zero size message */


  if ( buf_size > MAX_MSG ) bin_id = NBINS - 1; /* The last bin is for
                                                   messages > 1024MB */
  else if ( buf_size > 512*ONE_MB ) bin_id = sz_513to1024M;
  else if ( buf_size > 256*ONE_MB ) bin_id = sz_257to512M;
  else if ( buf_size > 128*ONE_MB ) bin_id = sz_129to256M;
  else if ( buf_size >  64*ONE_MB ) bin_id = sz_65to128M;
  else if ( buf_size >  32*ONE_MB ) bin_id = sz_33to64M;
  else if ( buf_size >  16*ONE_MB ) bin_id = sz_17to32M;
  else if ( buf_size >   8*ONE_MB ) bin_id = sz_9to16M;
  else if ( buf_size >   4*ONE_MB ) bin_id = sz_5to8M;
  else if ( buf_size >     ONE_MB ) bin_id = sz_1to4M;
  else if ( buf_size > 512*ONE_KB ) bin_id = sz_513to1024K;
  else if ( buf_size > 256*ONE_KB ) bin_id = sz_257to512K;
  else if ( buf_size > 128*ONE_KB ) bin_id = sz_129to256K;
  else if ( buf_size >  64*ONE_KB ) bin_id = sz_65to128K;
  else if ( buf_size >  32*ONE_KB ) bin_id = sz_33to64K;
  else if ( buf_size >  16*ONE_KB ) bin_id = sz_17to32K;
  else if ( buf_size >   8*ONE_KB ) bin_id = sz_9to16K;
  else if ( buf_size >   4*ONE_KB ) bin_id = sz_5to8K;
  else if ( buf_size >     ONE_KB ) bin_id = sz_1to4K;
  else if ( buf_size > 512 )        bin_id = sz_513to1024;
  else if ( buf_size > 256 )        bin_id = sz_257to512;
  else if ( buf_size > 128 )        bin_id = sz_129to256;
  else if ( buf_size >  64 )        bin_id = sz_65to128;
  else if ( buf_size >  32 )        bin_id = sz_33to64;
  else if ( buf_size >  16 )        bin_id = sz_17to32;
  else if ( buf_size >   8 )        bin_id = sz_9to16;
  else if ( buf_size >   4 )        bin_id = sz_5to8;
  else if ( buf_size >   0  )       bin_id = sz_1to4;

  return bin_id;
}

/*
    MPI_Finalize - prototyping replacement for MPI_Finalize
    Trace the beginning and ending of MPI_Finalize.
*/
int MPI_Finalize(  )
{
  int returnVal;
  int llrank, nprocs;
  char profname[100];
  FILE *pf;
  struct rusage myusage;
  int max_wall_clock_rank, min_wall_clock_rank;
  double wall_clock_t;
  double max_wall_clock_t, min_wall_clock_t;
  int max_user_rank, min_user_rank;
  double user_t;
  double max_user_t, min_user_t;
  int max_sys_rank, min_sys_rank;
  double sys_t;
  double max_sys_t, min_sys_t;

  long maxrss, idrss;
  long minflt, majflt;
  long nswap; 
  long inblock, oublock;
  long msgsnd, msgrcv; 
  long nsignals; 
  long nvcsw, nivcsw;

  double t_tmp;

  int ierr;

#ifdef HAVE_PCL
  int n_instr, n_cycles, n_fp_ops;
  double flops;
#endif

#ifdef GETRUSAGE_HAS_MAXRSS
  long maxrss_all, minrss_all, sumrss_all, rss_tmp;
#endif

#ifdef LINUX
  procstat_t stats;
  long maxrss_all, minrss_all, sumrss_all, rss_tmp;
#endif

  PMPI_Comm_rank( MPI_COMM_WORLD, &llrank );
  PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );

  ierr = PMPI_Barrier( MPI_COMM_WORLD );
  MPI_End_time = clock();
  wall_clock_t = ((double) MPI_End_time - MPI_Start_time)/CLOCKS_PER_SEC;

  /* Get Resource usage */
  if ( getrusage(RUSAGE_SELF, &myusage) < 0 )
    fprintf(stderr, "[%d] getrusage error\n", llrank);
  user_t   = (double) myusage.ru_utime.tv_sec +
              myusage.ru_utime.tv_usec / 1000000.0;
  sys_t	   = (double) myusage.ru_stime.tv_sec +
              myusage.ru_stime.tv_usec / 1000000.0;
  maxrss   = myusage.ru_maxrss;

  idrss	   = myusage.ru_idrss;
  minflt   = myusage.ru_minflt;
  majflt   = myusage.ru_majflt;
  nswap	   = myusage.ru_nswap;
  inblock  = myusage.ru_inblock;
  oublock  = myusage.ru_oublock;
  msgsnd   = myusage.ru_msgsnd;
  msgrcv   = myusage.ru_msgrcv;
  nsignals = myusage.ru_nsignals;
  nvcsw	   = myusage.ru_nvcsw;
  nivcsw   = myusage.ru_nivcsw;


  t_tmp = wall_clock_t;
  getrange_double( &t_tmp, 1, &wall_clock_t, &min_wall_clock_t, 
		   &max_wall_clock_t, &min_wall_clock_rank, 
		   &max_wall_clock_rank );
  t_tmp = user_t;
  getrange_double( &t_tmp, 1, &user_t, &min_user_t, &max_user_t, 
		   &min_user_rank, & max_user_rank );
  t_tmp = sys_t;
  getrange_double( &t_tmp, 1, &sys_t, &min_sys_t, &max_sys_t, 
		   &min_sys_rank, &max_sys_rank );

#ifdef GETRUSAGE_HAS_MAXRSS
  PMPI_Reduce( &maxrss, &maxrss_all, 1, MPI_LONG, MPI_MAX, 0, MPI_COMM_WORLD);
  PMPI_Reduce( &maxrss, &minrss_all, 1, MPI_LONG, MPI_MIN, 0, MPI_COMM_WORLD);
#endif

#ifdef LINUX
  ierr = ReadProcStat( &stats );
  maxrss = stats.rss;
  PMPI_Reduce( &maxrss, &maxrss_all, 1, MPI_LONG, MPI_MAX, 0, MPI_COMM_WORLD);
  PMPI_Reduce( &maxrss, &minrss_all, 1, MPI_LONG, MPI_MIN, 0, MPI_COMM_WORLD);
#endif

#define PRINT_HEADER(name) printout_header( name, pf, llrank )

  if ( llrank == 0 ) {
    if (getenv("MPI_PROFILE_FILE") != NULL) {
      sprintf(profname, "%s", getenv("MPI_PROFILE_FILE"));}
    else sprintf(profname,"%s", MPI_PROF_FILE_NAME);
    printf("Profile will be saved in %s.\n", profname);
    pf = fopen( profname, "w" );

    if (cmdname) {
	fprintf( pf, "Program %s\n", cmdname );
    }
    {
	time_t t;
	char *c;
	t = time(NULL);
	c = asctime(localtime(&t));
	fprintf(pf, "MPI_Finalize called at:\t%s\n", c);
	fprintf(pf, "Number of processors:\t%d\n\n", nprocs);
    }

    wall_clock_t /= nprocs;
    user_t	 /= nprocs;
    sys_t	 /= nprocs;

    fprintf(pf, "Timing Stats: [seconds] [min/max] [min rank/max rank]\n");
    fprintf(pf, "  wall-clock: %.4g sec   %f / %f    %d / %d\n",
	    wall_clock_t, min_wall_clock_t, max_wall_clock_t,
	    min_wall_clock_rank, max_wall_clock_rank);
    fprintf(pf, "        user: %.4g sec   %f / %f    %d / %d\n",
	    user_t, min_user_t, max_user_t, min_user_rank, max_user_rank);
    fprintf(pf, "         sys: %.4g sec   %f / %f    %d / %d\n",
	    sys_t, min_sys_t, max_sys_t, min_sys_rank, max_sys_rank);

#ifdef GETRUSAGE_HAS_MAXRSS
    fprintf(pf, "\n");
    fprintf(pf, "Memory Usage Stats (RSS) [min/max KB]:\t%ld/%ld\n",
    minrss_all, maxrss_all);
#endif

#ifdef LINUX
    fprintf(pf, "\n");
    fprintf(pf, "Memory Usage Stats (RSS) [min/max KB]:\t%ld/%ld\n",
    minrss_all, maxrss_all);
#endif

    PRINT_HEADER("MPI Routine Statistics");
  }

  PRINT_HEADER("Barriers and Waits");

#define PRINT_BAR_STATS(sname,name) \
  printout_barstats( sname##_calls, sname##_time, name, pf, llrank, nprocs )

  PRINT_BAR_STATS( barrier, "MPI_Barrier" );
  PRINT_BAR_STATS( wait, "MPI_Wait" );
  PRINT_BAR_STATS( waitall, "MPI_Waitall" );
  PRINT_BAR_STATS( waitany, "MPI_Waitany" );
  PRINT_BAR_STATS( waitsome, "MPI_Waitsome" );

  PRINT_HEADER("Message Routines");

  if ( llrank == 0 )
    fprintf(pf, "[max bin limit]  [min/max/avg calls] [min rank/max rank]  [min/max/avg msg size] [min rank/max rank] [min/max/avg time] [min rank/max rank]\n");

#define PRINT_MSG_STATS(sname,name) \
printout_msgstats_coll( sname##_calls, sname##_size, sname##_time, \
		   (void *)0, name, pf, llrank, nprocs )

#define PRINT_MSG_STATS_COLL(sname,name) \
printout_msgstats_coll( sname##_calls, sname##_size, sname##_time, \
		   sname##_timesync, name, pf, llrank, nprocs )

   PRINT_MSG_STATS_COLL( allgather, "MPI_Allgather" );
   PRINT_MSG_STATS_COLL( allgatherv, "MPI_Allgatherv" );
   PRINT_MSG_STATS_COLL( allreduce, "MPI_Allreduce" );
   PRINT_MSG_STATS_COLL( alltoall, "MPI_Alltoall" );
   PRINT_MSG_STATS_COLL( alltoallv, "MPI_Alltoallv" );
   PRINT_MSG_STATS( bcast, "MPI_Bcast" );
   PRINT_MSG_STATS( gather, "MPI_Gather" );
   PRINT_MSG_STATS( gatherv, "MPI_Gatherv" );
   PRINT_MSG_STATS( reduce_scatter, "MPI_Reduce_scatter" );
   PRINT_MSG_STATS( reduce, "MPI_Reduce" );
   PRINT_MSG_STATS( scan, "MPI_Scan" );
   PRINT_MSG_STATS( scatter, "MPI_Scatter" );
   PRINT_MSG_STATS( scatterv, "MPI_Scatterv" );

   PRINT_MSG_STATS( send, "MPI_Send" );
   PRINT_MSG_STATS( sendrecv, "MPI_Sendrecv" );
   PRINT_MSG_STATS( sendrecv_replace, "MPI_Sendrecv_replace" );
   PRINT_MSG_STATS( bsend, "MPI_Bsend" );
   PRINT_MSG_STATS( ibsend, "MPI_Ibsend" );
   PRINT_MSG_STATS( isend, "MPI_Isend" );
   PRINT_MSG_STATS( issend, "MPI_Issend" );
   PRINT_MSG_STATS( irsend, "MPI_Irsend" );
   PRINT_MSG_STATS( rsend, "MPI_Rsend" );
   PRINT_MSG_STATS( ssend, "MPI_Ssend" );
   PRINT_MSG_STATS( recv, "MPI_Recv" );
   PRINT_MSG_STATS( irecv, "MPI_Irecv" );
   PRINT_MSG_STATS( pack, "MPI_Pack" );
   PRINT_MSG_STATS( unpack, "MPI_Unpack" );

   PRINT_HEADER("Message Totals");
   printout_total_msgstats(pf, llrank, nprocs);
#ifdef HAVE_LIBPERFEX
   /* We can't use the print_costs routine because the output is not reliably
      written to the output (a known bug in the SGI MPI implementation) */
   if ( hwpc_ok ) {
     if ( llrank == 0 ) {
       if ( read_counters( hwpc_e0, &hwpc_c0, hwpc_e1, &hwpc_c1 ) < 0 ) {
	 fprintf(pf, "** Error reading hardware performance counters **\n");
       }
       else {
	 fprintf(pf,"\n------   Hardware Performance Stats\n");
	 /*print_counters(hwpc_e0, hwpc_c0, hwpc_e1, hwpc_c1);*/
	 /* print_costs(hwpc_e0, hwpc_c0, hwpc_e1, hwpc_c1); */
	 fprintf( pf, "Event %d \t%lld\n", hwpc_e0, hwpc_c0 );
	 fprintf( pf, "Event %d \t%lld\n", hwpc_e1, hwpc_c1 );
       }
     }
   }
   else {
     fprintf(pf, "** Error initializing hardware performance counters **\n");
   }
#endif
#ifdef HAVE_PCL
   /* Stop PCL measurements */
  if (PCLstop(i_result_list, fp_result_list, ncounter) != PCL_SUCCESS)
    fprintf(stderr, "[%d] PCLstop error!\n", llrank);

  n_fp_ops = i_result_list[1];
  n_cycles = i_result_list[0];
  flops    = fp_result_list[2];
  printout_pclstats(n_cycles, n_fp_ops, flops, pf, llrank, nprocs);
#endif
 
#define PRINT_COMM_DESTS(sname,name) \
  printout_comm_dests( sname##_target, name, pf )

  PRINT_HEADER("Communication Partners");
  PRINT_COMM_DESTS( send, "MPI_Send" );
  PRINT_COMM_DESTS( sendrecv, "MPI_Sendrecv" );
  PRINT_COMM_DESTS( sendrecv_replace, "MPI_Sendrecv_replace" );
  PRINT_COMM_DESTS( bsend, "MPI_Bsend" );
  PRINT_COMM_DESTS( ibsend, "MPI_Ibsend" );
  PRINT_COMM_DESTS( isend, "MPI_Isend" );
  PRINT_COMM_DESTS( issend, "MPI_Issend" );
  PRINT_COMM_DESTS( irsend, "MPI_Irsend" );
  PRINT_COMM_DESTS( rsend, "MPI_Rsend" );
  PRINT_COMM_DESTS( ssend, "MPI_Ssend" );
  PRINT_COMM_DESTS( recv, "MPI_Recv" );
  PRINT_COMM_DESTS( irecv, "MPI_Irecv" );

  if ( llrank == 0 ) fclose(pf);
  returnVal = PMPI_Finalize();
  return returnVal;
}

static void printout_msgstats_coll( int calls[], long sizes[], double times[],
				    double timesync[], 
				    const char *name, FILE *pf, int llrank, 
				    int nprocs )
{
  int sumcalls[NBINS], mincalls[NBINS], maxcalls[NBINS], 
      minloc[NBINS], maxloc[NBINS];
  int minsizeloc[NBINS], maxsizeloc[NBINS];
  long sumsizes[NBINS], minsizes[NBINS], maxsizes[NBINS];
  int mintimeloc[NBINS], maxtimeloc[NBINS];
  double sumtimes[NBINS], mintimes[NBINS], maxtimes[NBINS];
  double sumsynctimes[NBINS], minsynctimes[NBINS], maxsynctimes[NBINS];
  int mintimesyncloc[NBINS], maxtimesyncloc[NBINS];

  /* These are collective calls that reduce data to process 0 */
  getrange_int( calls, NBINS, sumcalls, mincalls, maxcalls, minloc, maxloc );
  getrange_long( sizes, NBINS, 
		 sumsizes, minsizes, maxsizes, minsizeloc, maxsizeloc );
  getrange_double( times, NBINS, 
		   sumtimes, mintimes, maxtimes, mintimeloc, maxtimeloc );
  if (timesync) {
      getrange_double( timesync, NBINS, 
		       sumsynctimes, minsynctimes, maxsynctimes, 
		       mintimesyncloc, maxtimesyncloc );
  }
  if (llrank == 0) {
      int i;
      for (i=0; i<NBINS; i++) {
	  if (sumcalls[i] > 0) {
#if defined(NCSAPERF)
	    fprintf( pf, "%-20s\t", name );
	    fprintf( pf, "%s\t%d\t%d\t%d\t%d\t%d\t", 
		     max_bin_names[i], mincalls[i], maxcalls[i], 
		     sumcalls[i]/nprocs, minloc[i], maxloc[i] );
	    fprintf( pf, "%ld\t%ld\t%ld\t%d\t%d\t ",
		     minsizes[i], maxsizes[i], sumsizes[i]/nprocs,
		     minsizeloc[i], maxsizeloc[i] );
	    fprintf( pf, "%f\t%f\t%f\t%d\t%d\t",
		     mintimes[i], maxtimes[i], sumtimes[i]/nprocs,
		     mintimeloc[i], maxtimeloc[i] );
	    if (timesync) {
	      fprintf( pf, "%f\t%f\t%f\t%d\t%d\n",
		       minsynctimes[i], maxsynctimes[i], 
		       sumsynctimes[i]/nprocs,
		       mintimesyncloc[i], maxtimesyncloc[i] );
	    }
	    else {
	      fputs( "\n", pf );
	    }
#else
	    fprintf( pf, "%-20s: ", name );
	    fprintf( pf, "%s\t%d %d %d %d %d \t", 
		     max_bin_names[i], mincalls[i], maxcalls[i], 
		     sumcalls[i]/nprocs, minloc[i], maxloc[i] );
	    fprintf( pf, "%ld %ld %ld %d %d \t ",
		     minsizes[i], maxsizes[i], sumsizes[i]/nprocs,
		     minsizeloc[i], maxsizeloc[i] );
	    fprintf( pf, "%f %f %f %d %d \t",
		     mintimes[i], maxtimes[i], sumtimes[i]/nprocs,
		     mintimeloc[i], maxtimeloc[i] );
	    if (timesync) {
	      fprintf( pf, "%f %f %f %d %d\n",
		       minsynctimes[i], maxsynctimes[i], 
		       sumsynctimes[i]/nprocs,
		       mintimesyncloc[i], maxtimesyncloc[i] );
	    }
	    else {
	      fputs( "\n", pf );
	    }
#endif
	  }
      }
  }
}

/* For each rank, calculate the total number of communication partners */
/* Print the average across ranks.                                     */
static void printout_comm_dests( int32 target[], const char *name, 
				 FILE *pf )
{
  int i,j;
  int32 SelectBit;

  int32 AllCommDests[MAX_PROCBY32];  /* Store all accessed ranks */
  int nprocs, rank;
  int partners, max_partners, min_partners, sum_partners;
  int max_loc, min_loc;

  PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );
  PMPI_Comm_rank( MPI_COMM_WORLD, &rank );

  partners = 0;
  for (i=0;i<MAX_PROCBY32;i++) {
    SelectBit = 1;
    for (j=0;j<32;j++) {
      if Bit_Is_Set(target,SelectBit) partners+=1;
      SelectBit = SelectBit << 1;
    }
  }

  /* Get the min/max/avg across procs, also get min_loc/max_loc */
  getrange_int_single(partners, &max_partners, 
		      &max_loc, &min_partners, 
		      &min_loc, &sum_partners);

  if ( rank == 0 && sum_partners > 0 ) {
    fprintf( pf, "%-20s\t", name );
    fprintf( pf, "%d %d %.1f %d %d \n", 
	     min_partners, max_partners, 
	     (float) sum_partners/nprocs, min_loc, max_loc);
  }
}

/* 
   The following collective routines compute the 
   min/max/rank of min/rank of max 
   of an array of values, returning the values in supplied arrays.

   These are three essentially identical routines, except for the choice of 
   input datatype (int, long, or double) and the related structures and
   MPI datatypes. 
 */

static void getrange_int( const int vals[], int count, 
			  int sumout[], int minout[], int maxout[],
			  int minloc[], int maxloc[] )
{
    typedef struct { int val; int loc; } twoint;
    twoint *lvals, *gvals;
    int i, nprocs, rank;

    lvals = (twoint *)malloc( 2 * count * sizeof(twoint) );
    gvals = lvals + count;

    PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );
    PMPI_Comm_rank( MPI_COMM_WORLD, &rank );
    for (i=0; i<count; i++) {
	lvals[i].val = vals[i];
	lvals[i].loc = rank;
    }
    PMPI_Reduce( lvals, gvals, count, MPI_2INT, MPI_MAXLOC, 
		 0, MPI_COMM_WORLD );
    for (i=0; i<count; i++) {
	maxout[i] = gvals[i].val;
	maxloc[i] = gvals[i].loc;
    }
    PMPI_Reduce( lvals, gvals, count, MPI_2INT, MPI_MINLOC, 
		 0, MPI_COMM_WORLD );
    for (i=0; i<count; i++) {
	minout[i] = gvals[i].val;
	minloc[i] = gvals[i].loc;
    }
    PMPI_Reduce( (int *)vals, sumout, count, MPI_INT, MPI_SUM, 
		 0, MPI_COMM_WORLD );
    free( lvals );
}

static void getrange_long( const long vals[], int count, 
			   long sumout[], long minout[], long maxout[], 
			   int minloc[], int maxloc[] )
{
    typedef struct { long val; int loc; } longint;
    longint *lvals, *gvals;
    int i, nprocs, rank;

    lvals = (longint *)malloc( 2 * count * sizeof(longint) );
    gvals = lvals + count;

    PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );
    PMPI_Comm_rank( MPI_COMM_WORLD, &rank );
    for (i=0; i<count; i++) {
	lvals[i].val = vals[i];
	lvals[i].loc = rank;
    }
    PMPI_Reduce( lvals, gvals, count, MPI_LONG_INT, MPI_MAXLOC, 
		 0, MPI_COMM_WORLD );
    for (i=0; i<count; i++) {
	maxout[i] = gvals[i].val;
	maxloc[i] = gvals[i].loc;
    }
    PMPI_Reduce( lvals, gvals, count, MPI_LONG_INT, MPI_MINLOC, 
		 0, MPI_COMM_WORLD );
    for (i=0; i<count; i++) {
	minout[i] = gvals[i].val;
	minloc[i] = gvals[i].loc;
    }
    PMPI_Reduce( (long *)vals, sumout, count, MPI_LONG, MPI_SUM, 
		 0, MPI_COMM_WORLD );
    free( lvals );
}

static void getrange_double( const double vals[], int count, 
			   double sumout[], double minout[], double maxout[], 
			     int minloc[], int maxloc[] )
{
    typedef struct { double val; int loc; } doubleint;
    doubleint *lvals, *gvals;
    int i, nprocs, rank;

    lvals = (doubleint *)malloc( 2 * count * sizeof(doubleint) );
    gvals = lvals + count;

    PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );
    PMPI_Comm_rank( MPI_COMM_WORLD, &rank );
    for (i=0; i<count; i++) {
	lvals[i].val = vals[i];
	lvals[i].loc = rank;
    }
    PMPI_Reduce( lvals, gvals, count, MPI_DOUBLE_INT, MPI_MAXLOC, 
		 0, MPI_COMM_WORLD );
    for (i=0; i<count; i++) {
	maxout[i] = gvals[i].val;
	maxloc[i] = gvals[i].loc;
    }
    PMPI_Reduce( lvals, gvals, count, MPI_DOUBLE_INT, MPI_MINLOC, 
		 0, MPI_COMM_WORLD );
    for (i=0; i<count; i++) {
	minout[i] = gvals[i].val;
	minloc[i] = gvals[i].loc;
    }
    PMPI_Reduce( (double *)vals, sumout, count, MPI_DOUBLE, MPI_SUM, 
		 0, MPI_COMM_WORLD );
    free( lvals );
}

/* Determine the min/max/sum of single integer values across procs */
static void getrange_int_single(int val, int *maxout, int *maxloc, 
			     int *minout, int *minloc, int *sumout)
{
  typedef struct { int val; int loc; } twoint;
  twoint *lvals, *gvals;
  int nprocs, rank;

  lvals = (twoint *) malloc( 2 * sizeof(twoint) );
  gvals = lvals + 1;
  
  PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );
  PMPI_Comm_rank( MPI_COMM_WORLD, &rank );

  lvals->val = val;
  lvals->loc = rank;

  PMPI_Reduce( lvals, gvals, 1, MPI_2INT, MPI_MAXLOC, 
	       0, MPI_COMM_WORLD );
  *maxout = gvals->val;
  *maxloc = gvals->loc;

  PMPI_Reduce( lvals, gvals, 1, MPI_2INT, MPI_MINLOC, 
	       0, MPI_COMM_WORLD );
  *minout = gvals->val;
  *minloc = gvals->loc;

  PMPI_Reduce( &val, sumout, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD );
  free( lvals );
}

static void getrange_long_single(long val, long *maxout, int *maxloc, 
			     long *minout, int *minloc, long *sumout)
{
  typedef struct { long val; int loc; } longint;
  longint *lvals, *gvals;
  int nprocs, rank;

  lvals = (longint *) malloc( 2 * sizeof(longint) );
  gvals = lvals + 1;
  
  PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );
  PMPI_Comm_rank( MPI_COMM_WORLD, &rank );

  lvals->val = val;
  lvals->loc = rank;

  PMPI_Reduce( lvals, gvals, 1, MPI_LONG_INT, MPI_MAXLOC, 
	       0, MPI_COMM_WORLD );
  *maxout = gvals->val;
  *maxloc = gvals->loc;

  PMPI_Reduce( lvals, gvals, 1, MPI_LONG_INT, MPI_MINLOC, 
	       0, MPI_COMM_WORLD );
  *minout = gvals->val;
  *minloc = gvals->loc;

  PMPI_Reduce( &val, sumout, 1, MPI_LONG, MPI_SUM, 0, MPI_COMM_WORLD );
  free( lvals );
}

static void getrange_double_single(double val, double *maxout, int *maxloc, 
			     double *minout, int *minloc, double *sumout)
{
  typedef struct { double val; int loc; } doubleint;
  doubleint *lvals, *gvals;
  int nprocs, rank;

  lvals = (doubleint *) malloc( 2 * sizeof(doubleint) );
  gvals = lvals + 1;
  
  PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );
  PMPI_Comm_rank( MPI_COMM_WORLD, &rank );

  lvals->val = val;
  lvals->loc = rank;

  PMPI_Reduce( lvals, gvals, 1, MPI_DOUBLE_INT, MPI_MAXLOC, 
	       0, MPI_COMM_WORLD );
  *maxout = gvals->val;
  *maxloc = gvals->loc;

  PMPI_Reduce( lvals, gvals, 1, MPI_DOUBLE_INT, MPI_MINLOC, 
	       0, MPI_COMM_WORLD );
  *minout = gvals->val;
  *minloc = gvals->loc;

  PMPI_Reduce( &val, sumout, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );
  free( lvals );
}

static void printout_barstats( int calls, double times,
			       const char *name, FILE *pf,
			       int rank, int nprocs)
{
  int sum_calls, max_calls, min_calls, max_calls_loc, min_calls_loc;
  double sum_times, max_times, min_times;
  int max_times_loc, min_times_loc;

  getrange_int_single(calls, &max_calls, &max_calls_loc, 
		   &min_calls, &min_calls_loc, &sum_calls);
  getrange_double_single(times, &max_times, &max_times_loc, 
		   &min_times, &min_times_loc, &sum_times);

  if ( rank == 0 && sum_calls > 0 ) {
    fprintf( pf, "%-20s\t", name );
    fprintf( pf, "\t%d %d %d %d %d", min_calls, max_calls,
	     sum_calls/nprocs, min_calls_loc, max_calls_loc);
    fprintf( pf, "\t%f %f %f %d %d \n", min_times, max_times,
	     sum_times/nprocs, min_times_loc, max_times_loc);
  }
}

#ifdef HAVE_PCL
static void printout_pclstats( int n_cycles, int n_fp_ops, 
			       double flops, FILE *pf, int rank, int nprocs)
{
  int i;
  int sum_cycles, max_cycles, min_cycles, max_cycles_loc, min_cycles_loc;
  int sum_fp_ops, max_fp_ops, min_fp_ops, max_fp_ops_loc, min_fp_ops_loc;
  double sum_flops, max_flops, min_flops;
  int max_flops_loc, min_flops_loc;

  getrange_int_single(n_cycles, &max_cycles, &max_cycles_loc, 
		   &min_cycles, &min_cycles_loc, &sum_cycles);
  getrange_int_single(n_fp_ops, &max_fp_ops, &max_fp_ops_loc, 
		   &min_fp_ops, &min_fp_ops_loc, &sum_fp_ops);
  getrange_double_single(flops, &max_flops, &max_flops_loc, 
		   &min_flops, &min_flops_loc, &sum_flops);

  /*  for (i=0;i<nprocs;i++) {
    fprintf(pf, "[%d] reports %f Mflops.\n", i,
	    (double) i_result_list[1]/(double) i_result_list[0] * 500.0);
  }*/

  if ( rank == 0 ) {
    fprintf( pf, "\nPerformance Counter Stats\n");
    fprintf( pf, "Cycles:\t%d %d %d %d %d \n", min_cycles, max_cycles,
	     sum_cycles/nprocs, min_cycles_loc, max_cycles_loc);
    fprintf( pf, "Fp Ops:\t%d %d %d %d %d \n", min_fp_ops, max_fp_ops,
	     sum_fp_ops/nprocs, min_fp_ops_loc, max_fp_ops_loc);
    fprintf( pf, "FLOPS:\t%f %f %f %d %d \n", min_flops, max_flops,
	     sum_flops/nprocs, min_flops_loc, max_flops_loc);
  }
}
#endif

/* 
   For each node, sum the total amount of data sent and the time spent 
   in communication. Report the min/max/average for this value and
   the min/max/average communication rate as well
   
   Do the same for the "barrier" time in the collective routines, except 
   here only report the total time (no data moved).
   
   Variables beginning with total_ are totals of all BINS.
   Variables beginning with sum_ are sums across all procs.
   
   Blocking communication data is reported separtely from non-blocking.
*/

static void printout_total_msgstats(FILE *pf, int rank, int nprocs)
{
  int i;
  int  max_loc, min_loc;

  int  max_total_isend_calls, min_total_isend_calls, sum_total_isend_calls;
  long max_total_isend_size, min_total_isend_size, sum_total_isend_size;
  double max_total_isend_time, min_total_isend_time, sum_total_isend_time;
  double comm_isend_rate, max_comm_isend_rate, min_comm_isend_rate;
  double sum_comm_isend_rate;

  int  max_total_bsend_calls, min_total_bsend_calls, sum_total_bsend_calls;
  long max_total_bsend_size, min_total_bsend_size, sum_total_bsend_size;
  double max_total_bsend_time, min_total_bsend_time, sum_total_bsend_time;
  double comm_bsend_rate, max_comm_bsend_rate, min_comm_bsend_rate;
  double sum_comm_bsend_rate;

  int  max_total_col_send_calls, min_total_col_send_calls;
  int  sum_total_col_send_calls;
  long max_total_col_send_size, min_total_col_send_size;
  long sum_total_col_send_size;
  double max_total_col_send_time, min_total_col_send_time;
  double sum_total_col_send_time;
  double comm_col_rate, max_comm_col_rate, min_comm_col_rate;
  double sum_comm_col_rate;
  double total_col_timesync, min_total_col_timesync, max_total_col_timesync;
  double sum_total_col_timesync;

  for (i=0;i<NBINS;i++) {

    /* Blocking sends */
    total_bsend_calls += send_calls[i] + sendrecv_calls[i] 
      + sendrecv_replace_calls[i]  + bsend_calls[i]
      + rsend_calls[i] + ssend_calls[i];
    
    total_bsend_size += send_size[i] + sendrecv_size[i] 
      + sendrecv_replace_size[i]  + bsend_size[i]
      + rsend_size[i] + ssend_size[i];
    
    total_bsend_time += send_time[i] + sendrecv_time[i] 
      + sendrecv_replace_time[i]  + bsend_time[i]
      + rsend_time[i] + ssend_time[i];
    
    /* Immediate sends */
    total_isend_calls += ibsend_calls[i] + isend_calls[i] + issend_calls[i]
      + irsend_calls[i] + bcast_calls[i] + gather_calls[i] + gatherv_calls[i]
      + reduce_calls[i] + reduce_scatter_calls[i] + scan_calls[i]
      + scatter_calls[i] + scatterv_calls[i];

    total_isend_size += ibsend_size[i] + isend_size[i] + issend_size[i]
      + irsend_size[i] + bcast_size[i] + gather_size[i] + gatherv_size[i]
      + reduce_size[i] + reduce_scatter_size[i] + scan_size[i]
      + scatter_size[i] + scatterv_size[i];

    total_isend_time += ibsend_time[i] + isend_time[i] + issend_time[i]
      + irsend_time[i] + bcast_time[i] + gather_time[i] + gatherv_time[i]
      + reduce_time[i] + reduce_scatter_time[i] + scan_time[i]
      + scatter_time[i] + scatterv_time[i];

    /* Collective (including sync-time) */
    total_col_send_calls += allgather_calls[i] + allgatherv_calls[i] 
      + allreduce_calls[i] + alltoall_calls[i] + alltoallv_calls[i];
    
    total_col_send_size += allgather_size[i] + allgatherv_size[i] 
      + allreduce_size[i] + alltoall_size[i] + alltoallv_size[i];
    
    total_col_send_time += allgather_time[i] + allgatherv_time[i] 
      + allreduce_time[i] + alltoall_time[i] + alltoallv_time[i];
    
    total_col_timesync += allgather_timesync[i] + allgatherv_timesync[i]
      + allreduce_timesync[i] + alltoall_timesync[i] + alltoallv_timesync[i];

  }

  /* Point-to-point sends */

  /* Blocking */
  getrange_int_single(total_bsend_calls, &max_total_bsend_calls, 
		      &max_loc, &min_total_bsend_calls, 
		      &min_loc, &sum_total_bsend_calls);

  if ( rank == 0 && sum_total_bsend_calls > 0 )
    fprintf( pf, "Blocking send calls:\t%d %d %d %d %d \n", 
	     min_total_bsend_calls, max_total_bsend_calls, 
	     sum_total_bsend_calls/nprocs, min_loc, max_loc);

  getrange_long_single(total_bsend_size, &max_total_bsend_size, 
		       &max_loc, &min_total_bsend_size, 
		       &min_loc, &sum_total_bsend_size);

  if ( rank == 0 && sum_total_bsend_calls > 0 )
    fprintf( pf, "Blocking send size:\t%ld %ld %ld %d %d \n", 
	     min_total_bsend_size, max_total_bsend_size, 
	     sum_total_bsend_size/nprocs, min_loc, max_loc);

  getrange_double_single(total_bsend_time, &max_total_bsend_time, 
		       &max_loc, &min_total_bsend_time, 
		       &min_loc, &sum_total_bsend_time);
  
  if ( rank == 0 && sum_total_bsend_calls > 0 )
    fprintf( pf, "Blocking Send time:\t%f %f %f %d %d \n", 
	     min_total_bsend_time, max_total_bsend_time, 
	     sum_total_bsend_time/nprocs, min_loc, max_loc);

  comm_bsend_rate = (double) total_bsend_size/total_bsend_time;
  getrange_double_single(comm_bsend_rate, &max_comm_bsend_rate, 
			 &max_loc, &min_comm_bsend_rate, &min_loc,
			 &sum_comm_bsend_rate);

  if ( rank == 0 && sum_total_bsend_calls > 0 )
    fprintf(pf, "Blocking comm rate:\t%.0f %.0f %.0f %d %d\n", 
	    min_comm_bsend_rate, max_comm_bsend_rate, 
	    sum_comm_bsend_rate/nprocs, min_loc, max_loc);

  /* Immediate */
  getrange_int_single(total_isend_calls, &max_total_isend_calls, 
		      &max_loc, &min_total_isend_calls, 
		      &min_loc, &sum_total_isend_calls);

  if ( rank == 0 && sum_total_isend_calls > 0 )
    fprintf( pf, "Immediate send calls:\t%d %d %d %d %d \n", 
	     min_total_isend_calls, max_total_isend_calls, 
	     sum_total_isend_calls/nprocs, min_loc, max_loc);

  getrange_long_single(total_isend_size, &max_total_isend_size, 
		       &max_loc, &min_total_isend_size, 
		       &min_loc, &sum_total_isend_size);

  if ( rank == 0 && sum_total_isend_calls > 0 )
    fprintf( pf, "Immediate send size:\t%ld %ld %ld %d %d \n", 
	     min_total_isend_size, max_total_isend_size, 
	     sum_total_isend_size/nprocs, min_loc, max_loc);

  getrange_double_single(total_isend_time, &max_total_isend_time, 
		       &max_loc, &min_total_isend_time, 
		       &min_loc, &sum_total_isend_time);
  
  if ( rank == 0 && sum_total_isend_calls > 0 )
    fprintf( pf, "Immediate send time:\t%f %f %f %d %d \n", 
	     min_total_isend_time, max_total_isend_time, 
	     sum_total_isend_time/nprocs, min_loc, max_loc);

  comm_isend_rate = (double) total_isend_size/total_isend_time;
  getrange_double_single(comm_isend_rate, &max_comm_isend_rate, 
			 &max_loc, &min_comm_isend_rate, &min_loc,
			 &sum_comm_isend_rate);

  if ( rank == 0 && sum_total_isend_calls > 0 )
    fprintf(pf, "Immediate comm rate:\t%.0f %.0f %.0f %d %d\n", 
	    min_comm_isend_rate, max_comm_isend_rate, 
	    sum_comm_isend_rate/nprocs, min_loc, max_loc);

  /* Collective sends */
  getrange_int_single(total_col_send_calls, &max_total_col_send_calls, 
		      &max_loc, &min_total_col_send_calls, 
		      &min_loc, &sum_total_col_send_calls);

  if ( rank == 0 && sum_total_col_send_calls > 0 )
    fprintf( pf, "\nCollective Send calls:\t%d %d %d %d %d \n", 
	     min_total_col_send_calls, max_total_col_send_calls, 
	     sum_total_col_send_calls/nprocs, min_loc, max_loc);

  getrange_long_single(total_col_send_size, &max_total_col_send_size, 
		       &max_loc, &min_total_col_send_size, 
		       &min_loc, &sum_total_col_send_size);

  if ( rank == 0 && sum_total_col_send_calls > 0 )
    fprintf( pf, "Collective Send sizes:\t%ld %ld %ld %d %d \n", 
	     min_total_col_send_size, max_total_col_send_size, 
	     sum_total_col_send_size/nprocs, min_loc, max_loc);

  getrange_double_single(total_col_send_time, &max_total_col_send_time, 
			 &max_loc, &min_total_col_send_time, 
			 &min_loc, &sum_total_col_send_time);
  
  if ( rank == 0 && sum_total_col_send_calls > 0 )
    fprintf( pf, "Collective Send times:\t%f %f %f %d %d \n", 
	     min_total_col_send_time, max_total_col_send_time, 
	     sum_total_col_send_time/nprocs, min_loc, max_loc);

  comm_col_rate	= (double) total_col_send_size/total_col_send_time;
  getrange_double_single(comm_col_rate, &max_comm_col_rate, &max_loc,
			 &min_comm_col_rate, &min_loc,
			 &sum_comm_col_rate);
  if ( rank == 0 && sum_total_col_send_calls > 0 )
    fprintf(pf, "Collective Comm rate:\t%f %f %f %d %d\n", min_comm_col_rate,
	    max_comm_col_rate, sum_comm_col_rate/nprocs, min_loc, max_loc);

  /* Sync time - time spent waiting for other procs */
  getrange_double_single(total_col_timesync, &max_total_col_timesync, 
			 &max_loc, &min_total_col_timesync, 
			 &min_loc, &sum_total_col_timesync);
  
  if ( rank == 0 && sum_total_col_send_calls > 0 )
    fprintf( pf, "Collective sync times:\t%f %f %f %d %d \n", 
	     min_total_col_timesync, max_total_col_timesync, 
	     sum_total_col_timesync/nprocs, min_loc, max_loc);
}

static void printout_header( const char *name, FILE *pf, int rank )
{
  if ( rank == 0 )
    fprintf(pf, "\n------   %-40s ------\n", name);
}

#ifdef LINUX
/* 
   This subroutine takes a pointer to a procstat structure as input. It 
   then determines the process id and fills in the fields of the procstat 
   structure with info from the /proc/pid/stat file.
*/
#define GET_NEXT_WORD(name) word = strtok(name, " ")

#define WORD_NOT_NULL (word != NULL)

int ReadProcStat( procstat_t *pstats )
{
  int mypid;
  int nBytes;
  char fname[25];
  int fp;
  char buffer[BUFSIZ];
  char *word;

  mypid = getpid();
  sprintf(fname, "/proc/%d/stat", mypid);

  if ( (fp = open(fname, O_RDONLY) ) > -1 ) {
    nBytes = read(fp, buffer, BUFSIZ);
    close(fp);
  }
  else {
    fprintf(stderr, "Error reading procstat file %s.\n", fname);
    return -1;
  }
  /* Parse the buffer containing space-separated stat field values */
  if ( nBytes > 0 ) {
    GET_NEXT_WORD(buffer);
    if WORD_NOT_NULL pstats->pid = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->comm = word;
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->state = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->ppid = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->pgrp = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->session = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->tty = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->tpgid = atoi(word);
    GET_NEXT_WORD(NULL); 
    if WORD_NOT_NULL pstats->flags = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->minflt = strtoul(word, (char **) NULL, 10);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->cminflt = strtoul(word, (char **) NULL, 10);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->majflt = strtoul(word, (char **) NULL, 10);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->cmajflt = strtoul(word, (char **) NULL, 10);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->utime = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->stime = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->cutime = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->cstime = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->counter = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->priority = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->timeout = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->itrealvalue = strtoul(word, (char **) NULL, 10);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->starttime = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->vsize = strtoul(word, (char **) NULL, 10);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->rss = strtoul(word, (char **) NULL, 10);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->rlim = strtoul(word, (char **) NULL, 10);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->startcode = strtoul(word, (char **) NULL, 10);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->endcode = strtoul(word, (char **) NULL, 10);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->startstack = strtoul(word, (char **) NULL, 10);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->kstkesp = strtoul(word, (char **) NULL, 10);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->kstkeip = strtoul(word, (char **) NULL, 10);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->signal = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->blocked = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->sigignore = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->sigcatch = atoi(word);
    GET_NEXT_WORD(NULL);
    if WORD_NOT_NULL pstats->wchan = strtoul(word, (char **) NULL, 10);
  }
  else {
    fprintf(stderr, "No data read from %s.\n", fname);
    return -2;
  }
  return 0;
}
#endif

int32 UArraySum(int32 TheArray[], int count)
{
  int i;
  int32 TheSum=0;

  for (i=0;i<count;i++)
    TheSum += TheArray[i];

  return TheSum;
}
